use simcore_rs::{Time, SimContext, Facility, RandomVar, simulation};
use rand::{distributions::Uniform, rngs::SmallRng, SeedableRng, Rng};

/// Globally shared data.
struct Shared {
	joe: Facility,
	wait_time: RandomVar
}

// helper constants
const SIM_DURATION : Time = 60.0*24.0*7.0*3.0; // simulation runs over 3 weeks
const SEED_A       : u64  = 100000;
const SEED_S       : u64  = 200000;
const BENCH_SAMPLES: u64  = 100;

/// Customer process with access to the barber and a random processing delay.
struct Customer { delay: Time }

impl Customer {
	pub async fn actions(self, sim: SimContext<'_,Shared>) {
		// access the barber and record the time for the report
		let arrival_time = sim.now();
		sim.shared().joe.seize().await;
		sim.shared().wait_time.tabulate(sim.now() - arrival_time);
		
		// spend time
		sim.advance(self.delay).await;
		// release the barber
		sim.shared().joe.release();
	}
}

async fn barbershop(sim: SimContext<'_,Shared>, duration: Time) {
	// pseudo random number generators referenced from within the main process
	let mut rng_a = SmallRng::seed_from_u64(SEED_A);
	let mut rng_s = SmallRng::seed_from_u64(SEED_S);
	
	// activate a process to generate the customers
	sim.activate(async move {
		let dist_a = Uniform::new(12.0, 24.0);
		let dist_s = Uniform::new(12.0, 18.0);
		loop {
			sim.advance(rng_a.sample(dist_a)).await;
			if sim.now() >= SIM_DURATION { return; }
			sim.activate(Customer {
				delay: rng_s.sample(dist_s)
			}.actions(sim));
		}
	});
	
	// wait until the store closes
	sim.advance(duration).await;
	
	// finish processing the queue (no more customers arrive)
	sim.shared().joe.seize().await;
}

fn main() {
	let result = simulation(
		// global data
		Shared { joe: Facility::new(), wait_time: RandomVar::new() },
		// simulation entry point
		|sim| sim.process(barbershop(sim, SIM_DURATION))
	);
	
	// print statistics
	println!("wait_time: {:#.3}", result.wait_time);
}
