use simcore_rs::{Time, SimContext, Sender, Receiver, RandomVar, channel, simulation};
use rand::{rngs::SmallRng, SeedableRng, Rng};
use rand_distr::{Exp, Normal, Distribution};
use std::cell::RefCell;

const HARBOR_DISTANCE: Time  = 10.0;
const FERRY_TIMEOUT  : Time  = 5.0;
const FERRY_CAPACITY : usize = 5;

struct Shared {
	master_rng: RefCell<SmallRng>,
	ferry_cargo_len: RandomVar,
	ferry_load_time: RandomVar,
	car_wait_time: RandomVar,
}

#[derive(Debug)]
struct Car {
	arrival_time: Time,
	load_duration: Time
}

struct Pier {
	rng: SmallRng,
	landing_site: Sender<Car>
}

struct Ferry {
	cargo: Vec<Car>,
	timeout: Time,
	travel_time: Time,
	piers: Vec<Receiver<Car>>
}

impl Pier {
	async fn actions(mut self, sim: SimContext<'_,Shared>) {
		let arrival_delay = Exp::new(0.1).unwrap();
		let loading_delay = Normal::new(0.5, 0.2).unwrap();
		
		loop {
			sim.advance(arrival_delay.sample(&mut self.rng)).await;
			self.landing_site.send(Car {
				arrival_time: sim.now(),
				load_duration: loading_delay.sample_iter(&mut self.rng).find(|&val| val >= 0.0).unwrap()
			}).await.expect("no ferries in the simulation");
		}
	}
}

impl Ferry {
	async fn actions(mut self, sim: SimContext<'_,Shared>) {
		loop {
			for pier in self.piers.iter() {
				// unload the cars
				for car in self.cargo.drain(..) {
					sim.advance(car.load_duration).await;
				}
				
				let begin_loading = sim.now();
				
				// wait until new cars arrive or a timeout occurs
				while self.cargo.len() < self.cargo.capacity() {
					if let Some(car) = sim.select(
						pier.recv(),
						async { sim.advance(self.timeout).await; None }
					).await {
						// a car arrived in time
						sim.shared().car_wait_time.tabulate(sim.now() - car.arrival_time);
						sim.advance(car.load_duration).await;
						self.cargo.push(car);
					} else {
						// the timeout has been triggered
						break;
					}
				}
				
				sim.shared().ferry_load_time.tabulate(sim.now() - begin_loading);
				sim.shared().ferry_cargo_len.tabulate(self.cargo.len() as f64);
				
				// travel to the next harbor
				sim.advance(self.travel_time).await;
			}
		}
	}
}

async fn ferry(sim: SimContext<'_,Shared>, duration: Time, ferries: usize, harbors: usize) {
	let mut ports = Vec::with_capacity(harbors);
	
	// create all of the harbors
	for _ in 0..harbors {
		let (sx, rx) = channel();
		let harbor = Pier {
			rng: SmallRng::from_seed(sim.shared().master_rng.borrow_mut().gen()),
			landing_site: sx
		};
		sim.activate(harbor.actions(sim));
		ports.push(rx);
	}
	
	// create all of the ferries
	for i in 0..ferries {
		let ferry = Ferry {
			cargo: Vec::with_capacity(FERRY_CAPACITY),
			timeout: FERRY_TIMEOUT,
			travel_time: HARBOR_DISTANCE,
			piers: ports.iter().skip(i)
			            .chain(ports.iter().take(i)).cloned().collect()
		};
		sim.activate(ferry.actions(sim));
	}
	
	// await the end of the simulation
	sim.advance(duration).await;
	
	// take cars into account that weren't picked up by a ferry
	for port in ports {
		for _ in 0..port.len() {
			let car = port.recv().await.unwrap();
			sim.shared().car_wait_time.tabulate(sim.now() - car.arrival_time);
		}
	}
}

const SEED: u64 = 100000;
const FERRY_COUNT: usize = 2;
const HARBOR_COUNT: usize = 4;

fn main() {
	let result = simulation(
		// global data
		Shared {
			master_rng: RefCell::new(SmallRng::seed_from_u64(SEED)),
			ferry_cargo_len: RandomVar::new(),
			ferry_load_time: RandomVar::new(),
			car_wait_time: RandomVar::new(),
		},
		// simulation entry point
		|sim| sim.process(ferry(sim, 24.0*60.0*7.0, FERRY_COUNT, HARBOR_COUNT))
	);
	
	println!("Number of harbors: {}", HARBOR_COUNT);
	println!("Number of ferries: {}", FERRY_COUNT);
	println!("Car wait time: {:#.3}", result.car_wait_time);
	println!("Ferry cargo len: {:#.3}", result.ferry_cargo_len);
	println!("Ferry load time: {:#.3}", result.ferry_load_time);
}
