use simcore_rs::{SimContext, Time, RandomVar, Control, until, simulation};
use std::{rc::Rc, cell::{RefCell, Cell}};
use rand::{rngs::SmallRng, SeedableRng, Rng};
use rand_distr::{Exp, Normal, Uniform, Distribution};
use rayon::prelude::*;

struct Shared {
	master_rng: RefCell<SmallRng>,
	sim_duration: Cell<Time>
}

struct Table {
	forks: Vec<Control<bool>>,
	forks_held: Control<usize>,
	forks_awaited: Control<usize>
}

impl Table {
	fn new(forks: usize) -> Self {
		Table {
			forks: (0..forks).map(|_| Control::new(false)).collect(),
			forks_held: Control::default(),
			forks_awaited: Control::default()
		}
	}
	
	// blocks until the requested fork is available and acquires it
	async fn acquire_fork(&self, i: usize) {
		let num = i % self.forks.len();
		self.forks_awaited.set(self.forks_awaited.get() + 1);
		until(&self.forks[num], |fork| !fork.get()).await;
		self.forks[num].set(true);
		self.forks_awaited.set(self.forks_awaited.get() - 1);
		self.forks_held.set(self.forks_held.get() + 1);
	}
	
	// returns the requested fork to the table
	fn release_fork(&self, i: usize) {
		self.forks[i % self.forks.len()].set(false);
		self.forks_held.set(self.forks_held.get() - 1);
	}
}

struct Philosopher {
	table: Rc<Table>,
	seat: usize,
	rng: SmallRng
}

impl Philosopher {
	async fn actions(mut self, sim: SimContext<'_, Shared>) {
		let thinking_duration = Exp::new(1.0).unwrap();
		let artificial_delay = Uniform::new(0.1, 0.2);
		let eating_duration = Normal::new(0.5, 0.2).unwrap();
		
		loop {
			// spend some time pondering the nature of things
			sim.advance(thinking_duration.sample(&mut self.rng)).await;
			
			// acquire the first fork
			self.table.acquire_fork(self.seat).await;
			
			// introduce an artificial delay to leave room for deadlocks
			sim.advance(artificial_delay.sample(&mut self.rng)).await;
			
			// acquire the second fork
			self.table.acquire_fork(self.seat + 1).await;
			
			// spend some time eating
			sim.advance(
				eating_duration
					.sample_iter(&mut self.rng)
					.find(|&val| val >= 0.0)
					.unwrap()
			).await;
			
			// release the forks
			self.table.release_fork(self.seat + 1);
			self.table.release_fork(self.seat);
		}
	}
}

async fn philosophers(sim: SimContext<'_, Shared>, count: usize) {
	let table = Rc::new(Table::new(count));
	
	// create the philosopher-processes and seat them
	for i in 0..count {
		sim.activate(Philosopher {
			table: table.clone(),
			seat: i,
			rng: SmallRng::from_seed(sim.shared().master_rng.borrow_mut().gen())
		}.actions(sim));
	}
	
	// wait for the precise configuration indicating a deadlock
	// (we technically don't really have to do this, because deadlock-detection
	//  is automatic since it means that no processes are scheduled and the
	//  simulation ends; SLX doesn't have that ability, though)
	until(
		(&table.forks_held, &table.forks_awaited), 
		|(held, awaited)|
			held.get() == count && awaited.get() == count
	).await;
	
	// tabulate the current system time
	sim.shared().sim_duration.set(sim.now());
}

const SEED              :   u64 = 100000;
const EXPERIMENT_COUNT  :   u64 =    500;
const PHILOSOPHER_COUNT : usize =      5;

fn main() {
	let sim_duration = (1..=EXPERIMENT_COUNT).into_par_iter()
		.map(|i|
			simulation(
				// global data
				Shared {
					master_rng: RefCell::new(SmallRng::seed_from_u64(i*SEED)),
					sim_duration: Cell::default()
				},
				// simulation entry point
				|sim| sim.process(philosophers(sim, PHILOSOPHER_COUNT))
			).sim_duration.get()
		)
		.fold(
			|| RandomVar::new(),
			|var, duration| { var.tabulate(duration); var }
		)
		.reduce(
			|| RandomVar::new(),
			|var_a, var_b| { var_a.merge(&var_b); var_a }
		);
	
	println!("sim_duration: {:#}", sim_duration);
}
